@@ -19,8 +19,10 @@
 #ifndef flac__decode_h
 #define flac__decode_h
 
+#include "analyze.h"
+
 /* outfile == 0 => test only */
-int decode_wav(const char *infile, const char *outfile, bool analysis_mode, bool verbose, uint64 skip);
-int decode_raw(const char *infile, const char *outfile, bool analysis_mode, bool verbose, uint64 skip, bool is_big_endian, bool is_unsigned_samples);
+int decode_wav(const char *infile, const char *outfile, bool analysis_mode, analysis_options aopts, bool verbose, uint64 skip);
+int decode_raw(const char *infile, const char *outfile, bool analysis_mode, analysis_options aopts, bool verbose, uint64 skip, bool is_big_endian, bool is_unsigned_samples);
 
 #endif
