@@ -21,6 +21,8 @@
 #include <stdlib.h> /* for malloc() */
 #include <string.h> /* for memcpy(), memset() */
 #include "private/bitbuffer.h"
+#include "private/bitmath.h"
+#include "private/crc.h"
 
 static const unsigned FLAC__BITBUFFER_DEFAULT_CAPACITY = 65536; /* bytes */
 
@@ -33,15 +35,6 @@ static const unsigned FLAC__BITBUFFER_DEFAULT_CAPACITY = 65536; /* bytes */
 #endif
 #define max(x,y) ((x)>(y)?(x):(y))
 
-static unsigned ilog2_(unsigned v)
-{
-	unsigned l = 0;
-	assert(v > 0);
-	while(v >>= 1)
-		l++;
-	return l;
-}
-
 static bool bitbuffer_resize_(FLAC__BitBuffer *bb, unsigned new_capacity)
 {
 	byte *new_buffer;
@@ -151,10 +144,18 @@ bool FLAC__bitbuffer_init_from(FLAC__BitBuffer *bb, const byte buffer[], unsigne
 	}
 }
 
+void FLAC__bitbuffer_init_read_crc16(FLAC__BitBuffer *bb, uint16 seed)
+{
+	assert(bb != 0);
+
+	bb->read_crc16 = seed;
+}
+
 bool FLAC__bitbuffer_concatenate_aligned(FLAC__BitBuffer *dest, const FLAC__BitBuffer *src)
 {
-	static byte mask_[9] = { 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff };
+	static const byte mask_[9] = { 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff };
 	unsigned bits_to_add = src->total_bits - src->total_consumed_bits;
+
 	assert(dest != 0);
 	assert(src != 0);
 
@@ -223,6 +224,7 @@ bool FLAC__bitbuffer_clone(FLAC__BitBuffer *dest, const FLAC__BitBuffer *src)
 	dest->consumed_bytes = src->consumed_bytes;
 	dest->consumed_bits = src->consumed_bits;
 	dest->total_consumed_bits = src->total_consumed_bits;
+	dest->read_crc16 = src->read_crc16;
 	return true;
 }
 
@@ -254,7 +256,7 @@ bool FLAC__bitbuffer_write_zeroes(FLAC__BitBuffer *bb, unsigned bits)
 
 bool FLAC__bitbuffer_write_raw_uint32(FLAC__BitBuffer *bb, uint32 val, unsigned bits)
 {
-	static uint32 mask[] = {
+	static const uint32 mask[] = {
 		0,
 		0x00000001, 0x00000003, 0x00000007, 0x0000000F,
 		0x0000001F, 0x0000003F, 0x0000007F, 0x000000FF,
@@ -328,7 +330,7 @@ bool FLAC__bitbuffer_write_raw_int32(FLAC__BitBuffer *bb, int32 val, unsigned bi
 
 bool FLAC__bitbuffer_write_raw_uint64(FLAC__BitBuffer *bb, uint64 val, unsigned bits)
 {
-	static uint64 mask[] = {
+	static const uint64 mask[] = {
 		0,
 		0x0000000000000001, 0x0000000000000003, 0x0000000000000007, 0x000000000000000F,
 		0x000000000000001F, 0x000000000000003F, 0x000000000000007F, 0x00000000000000FF,
@@ -400,6 +402,19 @@ bool FLAC__bitbuffer_write_raw_int64(FLAC__BitBuffer *bb, int64 val, unsigned bi
 	return FLAC__bitbuffer_write_raw_uint64(bb, (uint64)val, bits);
 }
 
+bool FLAC__bitbuffer_write_unary_unsigned(FLAC__BitBuffer *bb, unsigned val)
+{
+	if(val < 32)
+		return FLAC__bitbuffer_write_raw_uint32(bb, 1, ++val);
+	else if(val < 64)
+		return FLAC__bitbuffer_write_raw_uint64(bb, 1, ++val);
+	else {
+		if(!FLAC__bitbuffer_write_zeroes(bb, val))
+			return false;
+		return FLAC__bitbuffer_write_raw_uint32(bb, 1, 1);
+	}
+}
+
 unsigned FLAC__bitbuffer_rice_bits(int val, unsigned parameter)
 {
 	unsigned msbs, uval;
@@ -436,7 +451,7 @@ unsigned FLAC__bitbuffer_golomb_bits_signed(int val, unsigned parameter)
 	else
 		uval = (unsigned)(val << 1);
 
-	k = ilog2_(parameter);
+	k = FLAC__bitmath_ilog2(parameter);
 	if(parameter == 1u<<k) {
 		assert(k <= 30);
 
@@ -464,7 +479,7 @@ unsigned FLAC__bitbuffer_golomb_bits_unsigned(unsigned uval, unsigned parameter)
 
 	assert(parameter > 0);
 
-	k = ilog2_(parameter);
+	k = FLAC__bitmath_ilog2(parameter);
 	if(parameter == 1u<<k) {
 		assert(k <= 30);
 
@@ -485,9 +500,126 @@ unsigned FLAC__bitbuffer_golomb_bits_unsigned(unsigned uval, unsigned parameter)
 	return bits;
 }
 
+bool FLAC__bitbuffer_write_symmetric_rice_signed(FLAC__BitBuffer *bb, int val, unsigned parameter)
+{
+	unsigned total_bits, interesting_bits, msbs;
+	uint32 pattern;
+
+	assert(bb != 0);
+	assert(bb->buffer != 0);
+	assert(parameter <= 31);
+
+	/* init pattern with the unary end bit and the sign bit */
+	if(val < 0) {
+		pattern = 3;
+		val = -val;
+	}
+	else
+		pattern = 2;
+
+	msbs = val >> parameter;
+	interesting_bits = 2 + parameter;
+	total_bits = interesting_bits + msbs;
+	pattern <<= parameter;
+	pattern |= (val & ((1<<parameter)-1)); /* the binary LSBs */
+
+	if(total_bits <= 32) {
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, total_bits))
+			return false;
+	}
+	else {
+		/* write the unary MSBs */
+		if(!FLAC__bitbuffer_write_zeroes(bb, msbs))
+			return false;
+		/* write the unary end bit, the sign bit, and binary LSBs */
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, interesting_bits))
+			return false;
+	}
+	return true;
+}
+
+bool FLAC__bitbuffer_write_symmetric_rice_signed_guarded(FLAC__BitBuffer *bb, int val, unsigned parameter, unsigned max_bits, bool *overflow)
+{
+	unsigned total_bits, interesting_bits, msbs;
+	uint32 pattern;
+
+	assert(bb != 0);
+	assert(bb->buffer != 0);
+	assert(parameter <= 31);
+
+	*overflow = false;
+
+	/* init pattern with the unary end bit and the sign bit */
+	if(val < 0) {
+		pattern = 3;
+		val = -val;
+	}
+	else
+		pattern = 2;
+
+	msbs = val >> parameter;
+	interesting_bits = 2 + parameter;
+	total_bits = interesting_bits + msbs;
+	pattern <<= parameter;
+	pattern |= (val & ((1<<parameter)-1)); /* the binary LSBs */
+
+	if(total_bits <= 32) {
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, total_bits))
+			return false;
+	}
+	else if(total_bits > max_bits) {
+		*overflow = true;
+		return true;
+	}
+	else {
+		/* write the unary MSBs */
+		if(!FLAC__bitbuffer_write_zeroes(bb, msbs))
+			return false;
+		/* write the unary end bit, the sign bit, and binary LSBs */
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, interesting_bits))
+			return false;
+	}
+	return true;
+}
+
+bool FLAC__bitbuffer_write_symmetric_rice_signed_escape(FLAC__BitBuffer *bb, int val, unsigned parameter)
+{
+	unsigned total_bits, val_bits;
+	uint32 pattern;
+
+	assert(bb != 0);
+	assert(bb->buffer != 0);
+	assert(parameter <= 31);
+
+	val_bits = FLAC__bitmath_silog2(val);
+	total_bits = 2 + parameter + 5 + val_bits;
+
+	if(total_bits <= 32) {
+		pattern = 3;
+		pattern <<= (parameter + 5);
+		pattern |= val_bits;
+		pattern <<= val_bits;
+		pattern |= (val & ((1 << val_bits) - 1));
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, total_bits))
+			return false;
+	}
+	else {
+		/* write the '-0' escape code first */
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, 3u << parameter, 2+parameter))
+			return false;
+		/* write the length */
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, val_bits, 5))
+			return false;
+		/* write the value */
+		if(!FLAC__bitbuffer_write_raw_int32(bb, val, val_bits))
+			return false;
+	}
+	return true;
+}
+
 bool FLAC__bitbuffer_write_rice_signed(FLAC__BitBuffer *bb, int val, unsigned parameter)
 {
-	unsigned bits, msbs, uval;
+	unsigned total_bits, interesting_bits, msbs, uval;
 	uint32 pattern;
 
 	assert(bb != 0);
@@ -505,12 +637,13 @@ bool FLAC__bitbuffer_write_rice_signed(FLAC__BitBuffer *bb, int val, unsigned pa
 		uval = (unsigned)(val << 1);
 
 	msbs = uval >> parameter;
-	bits = 1 + parameter + msbs;
+	interesting_bits = 1 + parameter;
+	total_bits = interesting_bits + msbs;
 	pattern = 1 << parameter; /* the unary end bit */
 	pattern |= (uval & ((1<<parameter)-1)); /* the binary LSBs */
 
-	if(bits <= 32) {
-		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, bits))
+	if(total_bits <= 32) {
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, total_bits))
 			return false;
 	}
 	else {
@@ -518,7 +651,7 @@ bool FLAC__bitbuffer_write_rice_signed(FLAC__BitBuffer *bb, int val, unsigned pa
 		if(!FLAC__bitbuffer_write_zeroes(bb, msbs))
 			return false;
 		/* write the unary end bit and binary LSBs */
-		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, parameter+1))
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, interesting_bits))
 			return false;
 	}
 	return true;
@@ -526,7 +659,7 @@ bool FLAC__bitbuffer_write_rice_signed(FLAC__BitBuffer *bb, int val, unsigned pa
 
 bool FLAC__bitbuffer_write_rice_signed_guarded(FLAC__BitBuffer *bb, int val, unsigned parameter, unsigned max_bits, bool *overflow)
 {
-	unsigned bits, msbs, uval;
+	unsigned total_bits, interesting_bits, msbs, uval;
 	uint32 pattern;
 
 	assert(bb != 0);
@@ -546,15 +679,16 @@ bool FLAC__bitbuffer_write_rice_signed_guarded(FLAC__BitBuffer *bb, int val, uns
 		uval = (unsigned)(val << 1);
 
 	msbs = uval >> parameter;
-	bits = 1 + parameter + msbs;
+	interesting_bits = 1 + parameter;
+	total_bits = interesting_bits + msbs;
 	pattern = 1 << parameter; /* the unary end bit */
 	pattern |= (uval & ((1<<parameter)-1)); /* the binary LSBs */
 
-	if(bits <= 32) {
-		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, bits))
+	if(total_bits <= 32) {
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, total_bits))
 			return false;
 	}
-	else if(bits > max_bits) {
+	else if(total_bits > max_bits) {
 		*overflow = true;
 		return true;
 	}
@@ -563,7 +697,7 @@ bool FLAC__bitbuffer_write_rice_signed_guarded(FLAC__BitBuffer *bb, int val, uns
 		if(!FLAC__bitbuffer_write_zeroes(bb, msbs))
 			return false;
 		/* write the unary end bit and binary LSBs */
-		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, parameter+1))
+		if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, interesting_bits))
 			return false;
 	}
 	return true;
@@ -571,7 +705,7 @@ bool FLAC__bitbuffer_write_rice_signed_guarded(FLAC__BitBuffer *bb, int val, uns
 
 bool FLAC__bitbuffer_write_golomb_signed(FLAC__BitBuffer *bb, int val, unsigned parameter)
 {
-	unsigned bits, msbs, uval;
+	unsigned total_bits, msbs, uval;
 	unsigned k;
 
 	assert(bb != 0);
@@ -588,19 +722,19 @@ bool FLAC__bitbuffer_write_golomb_signed(FLAC__BitBuffer *bb, int val, unsigned
 	else
 		uval = (unsigned)(val << 1);
 
-	k = ilog2_(parameter);
+	k = FLAC__bitmath_ilog2(parameter);
 	if(parameter == 1u<<k) {
 		unsigned pattern;
 
 		assert(k <= 30);
 
 		msbs = uval >> k;
-		bits = 1 + k + msbs;
+		total_bits = 1 + k + msbs;
 		pattern = 1 << k; /* the unary end bit */
 		pattern |= (uval & ((1u<<k)-1)); /* the binary LSBs */
 
-		if(bits <= 32) {
-			if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, bits))
+		if(total_bits <= 32) {
+			if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, total_bits))
 				return false;
 		}
 		else {
@@ -639,26 +773,26 @@ bool FLAC__bitbuffer_write_golomb_signed(FLAC__BitBuffer *bb, int val, unsigned
 
 bool FLAC__bitbuffer_write_golomb_unsigned(FLAC__BitBuffer *bb, unsigned uval, unsigned parameter)
 {
-	unsigned bits, msbs;
+	unsigned total_bits, msbs;
 	unsigned k;
 
 	assert(bb != 0);
 	assert(bb->buffer != 0);
 	assert(parameter > 0);
 
-	k = ilog2_(parameter);
+	k = FLAC__bitmath_ilog2(parameter);
 	if(parameter == 1u<<k) {
 		unsigned pattern;
 
 		assert(k <= 30);
 
 		msbs = uval >> k;
-		bits = 1 + k + msbs;
+		total_bits = 1 + k + msbs;
 		pattern = 1 << k; /* the unary end bit */
 		pattern |= (uval & ((1u<<k)-1)); /* the binary LSBs */
 
-		if(bits <= 32) {
-			if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, bits))
+		if(total_bits <= 32) {
+			if(!FLAC__bitbuffer_write_raw_uint32(bb, pattern, total_bits))
 				return false;
 		}
 		else {
@@ -807,7 +941,7 @@ bool FLAC__bitbuffer_zero_pad_to_byte_boundary(FLAC__BitBuffer *bb)
 
 bool FLAC__bitbuffer_peek_bit(FLAC__BitBuffer *bb, unsigned *val, bool (*read_callback)(byte buffer[], unsigned *bytes, void *client_data), void *client_data)
 {
-	static byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
+	static const byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
 
 	/* to avoid a drastic speed penalty we don't:
 	assert(bb != 0);
@@ -815,21 +949,21 @@ bool FLAC__bitbuffer_peek_bit(FLAC__BitBuffer *bb, unsigned *val, bool (*read_ca
 	assert(bb->bits == 0);
 	*/
 
-read_bit_:
-	if(bb->total_consumed_bits < bb->total_bits) {
-		*val = (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
-		return true;
-	}
-	else {
-		if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
-			return false;
-		goto read_bit_;
+	while(1) {
+		if(bb->total_consumed_bits < bb->total_bits) {
+			*val = (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
+			return true;
+		}
+		else {
+			if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
+				return false;
+		}
 	}
 }
 
 bool FLAC__bitbuffer_read_bit(FLAC__BitBuffer *bb, unsigned *val, bool (*read_callback)(byte buffer[], unsigned *bytes, void *client_data), void *client_data)
 {
-	static byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
+	static const byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
 
 	/* to avoid a drastic speed penalty we don't:
 	assert(bb != 0);
@@ -837,27 +971,28 @@ bool FLAC__bitbuffer_read_bit(FLAC__BitBuffer *bb, unsigned *val, bool (*read_ca
 	assert(bb->bits == 0);
 	*/
 
-read_bit_:
-	if(bb->total_consumed_bits < bb->total_bits) {
-		*val = (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
-		bb->consumed_bits++;
-		if(bb->consumed_bits == 8) {
-			bb->consumed_bytes++;
-			bb->consumed_bits = 0;
+	while(1) {
+		if(bb->total_consumed_bits < bb->total_bits) {
+			*val = (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
+			bb->consumed_bits++;
+			if(bb->consumed_bits == 8) {
+				FLAC__CRC16_UPDATE(bb->buffer[bb->consumed_bytes], bb->read_crc16);
+				bb->consumed_bytes++;
+				bb->consumed_bits = 0;
+			}
+			bb->total_consumed_bits++;
+			return true;
+		}
+		else {
+			if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
+				return false;
 		}
-		bb->total_consumed_bits++;
-		return true;
-	}
-	else {
-		if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
-			return false;
-		goto read_bit_;
 	}
 }
 
 bool FLAC__bitbuffer_read_bit_to_uint32(FLAC__BitBuffer *bb, uint32 *val, bool (*read_callback)(byte buffer[], unsigned *bytes, void *client_data), void *client_data)
 {
-	static byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
+	static const byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
 
 	/* to avoid a drastic speed penalty we don't:
 	assert(bb != 0);
@@ -865,28 +1000,29 @@ bool FLAC__bitbuffer_read_bit_to_uint32(FLAC__BitBuffer *bb, uint32 *val, bool (
 	assert(bb->bits == 0);
 	*/
 
-read_bit_:
-	if(bb->total_consumed_bits < bb->total_bits) {
-		*val <<= 1;
-		*val |= (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
-		bb->consumed_bits++;
-		if(bb->consumed_bits == 8) {
-			bb->consumed_bytes++;
-			bb->consumed_bits = 0;
+	while(1) {
+		if(bb->total_consumed_bits < bb->total_bits) {
+			*val <<= 1;
+			*val |= (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
+			bb->consumed_bits++;
+			if(bb->consumed_bits == 8) {
+				FLAC__CRC16_UPDATE(bb->buffer[bb->consumed_bytes], bb->read_crc16);
+				bb->consumed_bytes++;
+				bb->consumed_bits = 0;
+			}
+			bb->total_consumed_bits++;
+			return true;
+		}
+		else {
+			if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
+				return false;
 		}
-		bb->total_consumed_bits++;
-		return true;
-	}
-	else {
-		if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
-			return false;
-		goto read_bit_;
 	}
 }
 
 bool FLAC__bitbuffer_read_bit_to_uint64(FLAC__BitBuffer *bb, uint64 *val, bool (*read_callback)(byte buffer[], unsigned *bytes, void *client_data), void *client_data)
 {
-	static byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
+	static const byte mask[] = { 128, 64, 32, 16, 8, 4, 2, 1 };
 
 	/* to avoid a drastic speed penalty we don't:
 	assert(bb != 0);
@@ -894,22 +1030,23 @@ bool FLAC__bitbuffer_read_bit_to_uint64(FLAC__BitBuffer *bb, uint64 *val, bool (
 	assert(bb->bits == 0);
 	*/
 
-read_bit_:
-	if(bb->total_consumed_bits < bb->total_bits) {
-		*val <<= 1;
-		*val |= (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
-		bb->consumed_bits++;
-		if(bb->consumed_bits == 8) {
-			bb->consumed_bytes++;
-			bb->consumed_bits = 0;
+	while(1) {
+		if(bb->total_consumed_bits < bb->total_bits) {
+			*val <<= 1;
+			*val |= (bb->buffer[bb->consumed_bytes] & mask[bb->consumed_bits])? 1 : 0;
+			bb->consumed_bits++;
+			if(bb->consumed_bits == 8) {
+				FLAC__CRC16_UPDATE(bb->buffer[bb->consumed_bytes], bb->read_crc16);
+				bb->consumed_bytes++;
+				bb->consumed_bits = 0;
+			}
+			bb->total_consumed_bits++;
+			return true;
+		}
+		else {
+			if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
+				return false;
 		}
-		bb->total_consumed_bits++;
-		return true;
-	}
-	else {
-		if(!bitbuffer_read_from_client_(bb, read_callback, client_data))
-			return false;
-		goto read_bit_;
 	}
 }
 
@@ -1003,6 +1140,58 @@ bool FLAC__bitbuffer_read_raw_int64(FLAC__BitBuffer *bb, int64 *val, unsigned bi
 	return true;
 }
 
+bool FLAC__bitbuffer_read_unary_unsigned(FLAC__BitBuffer *bb, unsigned *val, bool (*read_callback)(byte buffer[], unsigned *bytes, void *client_data), void *client_data)
+{
+	unsigned bit, val_ = 0;
+
+	assert(bb != 0);
+	assert(bb->buffer != 0);
+
+	while(1) {
+		if(!FLAC__bitbuffer_read_bit(bb, &bit, read_callback, client_data))
+			return false;
+		if(bit)
+			break;
+		else
+			val_++;
+	}
+	*val = val_;
+	return true;
+}
+
+bool FLAC__bitbuffer_read_symmetric_rice_signed(FLAC__BitBuffer *bb, int *val, unsigned parameter, bool (*read_callback)(byte buffer[], unsigned *bytes, void *client_data), void *client_data)
+{
+	uint32 sign = 0, lsbs = 0, msbs = 0;
+	unsigned bit;
+
+	assert(bb != 0);
+	assert(bb->buffer != 0);
+	assert(parameter <= 31);
+
+	/* read the unary MSBs and end bit */
+	while(1) {
+		if(!FLAC__bitbuffer_read_bit(bb, &bit, read_callback, client_data))
+			return false;
+		if(bit)
+			break;
+		else
+			msbs++;
+	}
+	/* read the sign bit */
+	if(!FLAC__bitbuffer_read_bit_to_uint32(bb, &sign, read_callback, client_data))
+		return false;
+	/* read the binary LSBs */
+	if(!FLAC__bitbuffer_read_raw_uint32(bb, &lsbs, parameter, read_callback, client_data))
+		return false;
+
+	/* compose the value */
+	*val = (msbs << parameter) | lsbs;
+	if(sign)
+		*val = -(*val);
+
+	return true;
+}
+
 bool FLAC__bitbuffer_read_rice_signed(FLAC__BitBuffer *bb, int *val, unsigned parameter, bool (*read_callback)(byte buffer[], unsigned *bytes, void *client_data), void *client_data)
 {
 	uint32 lsbs = 0, msbs = 0;
@@ -1010,7 +1199,7 @@ bool FLAC__bitbuffer_read_rice_signed(FLAC__BitBuffer *bb, int *val, unsigned pa
 
 	assert(bb != 0);
 	assert(bb->buffer != 0);
-	assert(parameter <= 32);
+	assert(parameter <= 31);
 
 	/* read the unary MSBs and end bit */
 	while(1) {
@@ -1042,7 +1231,7 @@ bool FLAC__bitbuffer_read_golomb_signed(FLAC__BitBuffer *bb, int *val, unsigned
 	assert(bb != 0);
 	assert(bb->buffer != 0);
 
-	k = ilog2_(parameter);
+	k = FLAC__bitmath_ilog2(parameter);
 
 	/* read the unary MSBs and end bit */
 	while(1) {
@@ -1091,7 +1280,7 @@ bool FLAC__bitbuffer_read_golomb_unsigned(FLAC__BitBuffer *bb, unsigned *val, un
 	assert(bb != 0);
 	assert(bb->buffer != 0);
 
-	k = ilog2_(parameter);
+	k = FLAC__bitmath_ilog2(parameter);
 
 	/* read the unary MSBs and end bit */
 	while(1) {
@@ -1161,23 +1350,24 @@ bool FLAC__bitbuffer_read_utf8_uint32(FLAC__BitBuffer *bb, uint32 *val, bool (*r
 		v = x & 0x01;
 		i = 5;
 	}
-	else
-		goto invalid_;
+	else {
+		*val = 0xffffffff;
+		return true;
+	}
 	for( ; i; i--) {
 		if(!FLAC__bitbuffer_read_raw_uint32(bb, &x, 8, read_callback, client_data))
 			return false;
 		if(raw)
 			raw[(*rawlen)++] = (byte)x;
-		if(!(x & 0x80) || (x & 0x40)) /* 10xxxxxx */
-			goto invalid_;
+		if(!(x & 0x80) || (x & 0x40)) { /* 10xxxxxx */
+			*val = 0xffffffff;
+			return true;
+		}
 		v <<= 6;
 		v |= (x & 0x3F);
 	}
 	*val = v;
 	return true;
-invalid_:
-	*val = 0xffffffff;
-	return true;
 }
 
 /* on return, if *val == 0xffffffffffffffff then the utf-8 sequence was invalid, but the return value will be true */
@@ -1219,23 +1409,24 @@ bool FLAC__bitbuffer_read_utf8_uint64(FLAC__BitBuffer *bb, uint64 *val, bool (*r
 		v = 0;
 		i = 6;
 	}
-	else
-		goto invalid_;
+	else {
+		*val = 0xffffffffffffffff;
+		return true;
+	}
 	for( ; i; i--) {
 		if(!FLAC__bitbuffer_read_raw_uint32(bb, &x, 8, read_callback, client_data))
 			return false;
 		if(raw)
 			raw[(*rawlen)++] = (byte)x;
-		if(!(x & 0x80) || (x & 0x40)) /* 10xxxxxx */
-			goto invalid_;
+		if(!(x & 0x80) || (x & 0x40)) { /* 10xxxxxx */
+			*val = 0xffffffffffffffff;
+			return true;
+		}
 		v <<= 6;
 		v |= (x & 0x3F);
 	}
 	*val = v;
 	return true;
-invalid_:
-	*val = 0xffffffff;
-	return true;
 }
 
 void FLAC__bitbuffer_dump(const FLAC__BitBuffer *bb, FILE *out)
